@@ -2,9 +2,11 @@
 
 # Copyright (c) IPython Development Team.
 # Distributed under the terms of the Modified BSD License.
+from __future__ import annotations
 
 import builtins
 import sys
+import typing as t
 
 from IPython.core.displayhook import DisplayHook
 from jupyter_client.session import Session, extract_header
@@ -20,6 +22,7 @@ class ZMQDisplayHook:
     topic = b"execute_result"
 
     def __init__(self, session, pub_socket):
+        """Initialize the hook."""
         self.session = session
         self.pub_socket = pub_socket
         self.parent_header = {}
@@ -29,6 +32,7 @@ class ZMQDisplayHook:
         return 0
 
     def __call__(self, obj):
+        """Handle a hook call."""
         if obj is None:
             return
 
@@ -45,6 +49,7 @@ class ZMQDisplayHook:
         )
 
     def set_parent(self, parent):
+        """Set the parent header."""
         self.parent_header = extract_header(parent)
 
 
@@ -58,33 +63,39 @@ class ZMQShellDisplayHook(DisplayHook):
     session = Instance(Session, allow_none=True)
     pub_socket = Any(allow_none=True)
     parent_header = Dict({})
+    msg: dict[str, t.Any] | None
 
     def set_parent(self, parent):
         """Set the parent for outbound messages."""
         self.parent_header = extract_header(parent)
 
     def start_displayhook(self):
-        self.msg = self.session.msg(
-            "execute_result",
-            {
-                "data": {},
-                "metadata": {},
-            },
-            parent=self.parent_header,
-        )
+        """Start the display hook."""
+        if self.session:
+            self.msg = self.session.msg(
+                "execute_result",
+                {
+                    "data": {},
+                    "metadata": {},
+                },
+                parent=self.parent_header,
+            )
 
     def write_output_prompt(self):
         """Write the output prompt."""
-        self.msg["content"]["execution_count"] = self.prompt_count
+        if self.msg:
+            self.msg["content"]["execution_count"] = self.prompt_count
 
     def write_format_data(self, format_dict, md_dict=None):
-        self.msg["content"]["data"] = json_clean(encode_images(format_dict))
-        self.msg["content"]["metadata"] = md_dict
+        """Write format data to the message."""
+        if self.msg:
+            self.msg["content"]["data"] = json_clean(encode_images(format_dict))
+            self.msg["content"]["metadata"] = md_dict
 
     def finish_displayhook(self):
         """Finish up all displayhook activities."""
         sys.stdout.flush()
         sys.stderr.flush()
-        if self.msg["content"]["data"]:
+        if self.msg and self.msg["content"]["data"] and self.session:
             self.session.send(self.pub_socket, self.msg, ident=self.topic)
         self.msg = None
