@@ -9,7 +9,7 @@ import sys
 from functools import partial
 
 import zmq
-from packaging.version import Version as V  # noqa
+from packaging.version import Version as V
 from traitlets.config.application import Application
 
 
@@ -52,7 +52,7 @@ def register_integration(*toolkitnames):
         for name in toolkitnames:
             loop_map[name] = func
 
-        func.exit_hook = lambda kernel: None
+        func.exit_hook = lambda kernel: None  # noqa: ARG005
 
         def exit_decorator(exit_func):
             """@func.exit is now a decorator
@@ -82,6 +82,11 @@ def _notify_stream_qt(kernel):
         def enum_helper(name):
             return operator.attrgetter(name.rpartition(".")[0])(sys.modules[QtCore.__package__])
 
+    def exit_loop():
+        """fall back to main loop"""
+        kernel._qt_notifier.setEnabled(False)
+        kernel.app.qt_event_loop.quit()
+
     def process_stream_events():
         """fall back to main loop when there's a socket event"""
         # call flush to ensure that the stream doesn't lose events
@@ -89,18 +94,34 @@ def _notify_stream_qt(kernel):
         # flush returns the number of events consumed.
         # if there were any, wake it up
         if kernel.shell_stream.flush(limit=1):
-            kernel._qt_notifier.setEnabled(False)
-            kernel.app.qt_event_loop.quit()
+            exit_loop()
 
     if not hasattr(kernel, "_qt_notifier"):
         fd = kernel.shell_stream.getsockopt(zmq.FD)
         kernel._qt_notifier = QtCore.QSocketNotifier(
-            fd, enum_helper('QtCore.QSocketNotifier.Type').Read, kernel.app.qt_event_loop
+            fd, enum_helper("QtCore.QSocketNotifier.Type").Read, kernel.app.qt_event_loop
         )
         kernel._qt_notifier.activated.connect(process_stream_events)
     else:
         kernel._qt_notifier.setEnabled(True)
 
+    # allow for scheduling exits from the loop in case a timeout needs to
+    # be set from the kernel level
+    def _schedule_exit(delay):
+        """schedule fall back to main loop in [delay] seconds"""
+        # The signatures of QtCore.QTimer.singleShot are inconsistent between PySide and PyQt
+        # if setting the TimerType, so we create a timer explicitly and store it
+        # to avoid a memory leak.
+        # PreciseTimer is needed so we exit after _at least_ the specified delay, not within 5% of it
+        if not hasattr(kernel, "_qt_timer"):
+            kernel._qt_timer = QtCore.QTimer(kernel.app)
+            kernel._qt_timer.setSingleShot(True)
+            kernel._qt_timer.setTimerType(enum_helper("QtCore.Qt.TimerType").PreciseTimer)
+            kernel._qt_timer.timeout.connect(exit_loop)
+        kernel._qt_timer.start(int(1000 * delay))
+
+    loop_qt._schedule_exit = _schedule_exit
+
     # there may already be unprocessed events waiting.
     # these events will not wake zmq's edge-triggered FD
     # since edge-triggered notification only occurs on new i/o activity.
@@ -108,11 +129,7 @@ def _notify_stream_qt(kernel):
     # so we start in a clean state ensuring that any new i/o events will notify.
     # schedule first call on the eventloop as soon as it's running,
     # so we don't block here processing events
-    if not hasattr(kernel, "_qt_timer"):
-        kernel._qt_timer = QtCore.QTimer(kernel.app)
-        kernel._qt_timer.setSingleShot(True)
-        kernel._qt_timer.timeout.connect(process_stream_events)
-    kernel._qt_timer.start(0)
+    QtCore.QTimer.singleShot(0, process_stream_events)
 
 
 @register_integration("qt", "qt5", "qt6")
@@ -125,7 +142,7 @@ def loop_qt(kernel):
 
     # `exec` blocks until there's ZMQ activity.
     el = kernel.app.qt_event_loop  # for brevity
-    el.exec() if hasattr(el, 'exec') else el.exec_()
+    el.exec() if hasattr(el, "exec") else el.exec_()
     kernel.app._in_event_loop = False
 
 
@@ -168,7 +185,7 @@ def loop_wx(kernel):
 
     # We have to put the wx.Timer in a wx.Frame for it to fire properly.
     # We make the Frame hidden when we create it in the main app below.
-    class TimerFrame(wx.Frame):
+    class TimerFrame(wx.Frame):  # type:ignore[misc]
         def __init__(self, func):
             wx.Frame.__init__(self, None, -1)
             self.timer = wx.Timer(self)
@@ -182,7 +199,7 @@ def loop_wx(kernel):
 
     # We need a custom wx.App to create our Frame subclass that has the
     # wx.Timer to defer back to the tornado event loop.
-    class IPWxApp(wx.App):
+    class IPWxApp(wx.App):  # type:ignore[misc]
         def OnInit(self):
             self.frame = TimerFrame(wake)
             self.frame.Show(False)
@@ -229,23 +246,33 @@ def loop_tk(kernel):
                 self.app = app
                 self.app.withdraw()
 
-        def process_stream_events(stream, *a, **kw):
+        def exit_loop():
+            """fall back to main loop"""
+            app.tk.deletefilehandler(kernel.shell_stream.getsockopt(zmq.FD))
+            app.quit()
+            app.destroy()
+            del kernel.app_wrapper
+
+        def process_stream_events(*a, **kw):
             """fall back to main loop when there's a socket event"""
-            if stream.flush(limit=1):
-                app.tk.deletefilehandler(stream.getsockopt(zmq.FD))
-                app.quit()
-                app.destroy()
-                del kernel.app_wrapper
+            if kernel.shell_stream.flush(limit=1):
+                exit_loop()
+
+        # allow for scheduling exits from the loop in case a timeout needs to
+        # be set from the kernel level
+        def _schedule_exit(delay):
+            """schedule fall back to main loop in [delay] seconds"""
+            app.after(int(1000 * delay), exit_loop)
+
+        loop_tk._schedule_exit = _schedule_exit
 
         # For Tkinter, we create a Tk object and call its withdraw method.
         kernel.app_wrapper = BasicAppWrapper(app)
-
-        notifier = partial(process_stream_events, kernel.shell_stream)
-        # seems to be needed for tk
-        notifier.__name__ = "notifier"  # type:ignore[attr-defined]
-        app.tk.createfilehandler(kernel.shell_stream.getsockopt(zmq.FD), READABLE, notifier)
+        app.tk.createfilehandler(
+            kernel.shell_stream.getsockopt(zmq.FD), READABLE, process_stream_events
+        )
         # schedule initial call after start
-        app.after(0, notifier)
+        app.after(0, process_stream_events)
 
         app.mainloop()
 
@@ -429,7 +456,7 @@ def loop_asyncio_exit(kernel):
         close_loop()
 
     elif not loop.is_closed():
-        loop.run_until_complete(close_loop)  # type:ignore
+        loop.run_until_complete(close_loop)  # type:ignore[arg-type]
         loop.close()
 
 
@@ -465,16 +492,16 @@ def set_qt_api_env_from_gui(gui):
     loaded = loaded_api()
 
     qt_env2gui = {
-        QT_API_PYSIDE2: 'qt5',
-        QT_API_PYQT5: 'qt5',
-        QT_API_PYSIDE6: 'qt6',
-        QT_API_PYQT6: 'qt6',
+        QT_API_PYSIDE2: "qt5",
+        QT_API_PYQT5: "qt5",
+        QT_API_PYSIDE6: "qt6",
+        QT_API_PYQT6: "qt6",
     }
-    if loaded is not None and gui != 'qt' and qt_env2gui[loaded] != gui:
-        print(f'Cannot switch Qt versions for this session; you must use {qt_env2gui[loaded]}.')
+    if loaded is not None and gui != "qt" and qt_env2gui[loaded] != gui:
+        print(f"Cannot switch Qt versions for this session; you must use {qt_env2gui[loaded]}.")
         return
 
-    if qt_api is not None and gui != 'qt':
+    if qt_api is not None and gui != "qt":
         if qt_env2gui[qt_api] != gui:
             print(
                 f'Request for "{gui}" will be ignored because `QT_API` '
@@ -482,44 +509,44 @@ def set_qt_api_env_from_gui(gui):
             )
             return
     else:
-        if gui == 'qt5':
+        if gui == "qt5":
             try:
-                import PyQt5  # noqa
+                import PyQt5  # noqa: F401
 
                 os.environ["QT_API"] = "pyqt5"
             except ImportError:
                 try:
-                    import PySide2  # noqa
+                    import PySide2  # noqa: F401
 
                     os.environ["QT_API"] = "pyside2"
                 except ImportError:
                     os.environ["QT_API"] = "pyqt5"
-        elif gui == 'qt6':
+        elif gui == "qt6":
             try:
-                import PyQt6  # noqa
+                import PyQt6  # noqa: F401
 
                 os.environ["QT_API"] = "pyqt6"
             except ImportError:
                 try:
-                    import PySide6  # noqa
+                    import PySide6  # noqa: F401
 
                     os.environ["QT_API"] = "pyside6"
                 except ImportError:
                     os.environ["QT_API"] = "pyqt6"
-        elif gui == 'qt':
+        elif gui == "qt":
             # Don't set QT_API; let IPython logic choose the version.
-            if 'QT_API' in os.environ:
-                del os.environ['QT_API']
+            if "QT_API" in os.environ:
+                del os.environ["QT_API"]
         else:
             print(f'Unrecognized Qt version: {gui}. Should be "qt5", "qt6", or "qt".')
             return
 
     # Do the actual import now that the environment variable is set to make sure it works.
     try:
-        from IPython.external.qt_for_kernel import QtCore, QtGui  # noqa
+        pass
     except Exception as e:
         # Clear the environment variable for the next attempt.
-        if 'QT_API' in os.environ:
+        if "QT_API" in os.environ:
             del os.environ["QT_API"]
             print(f"QT_API couldn't be set due to error {e}")
         return
@@ -527,7 +554,7 @@ def set_qt_api_env_from_gui(gui):
 
 def make_qt_app_for_kernel(gui, kernel):
     """Sets the `QT_API` environment variable if it isn't already set."""
-    if hasattr(kernel, 'app'):
+    if hasattr(kernel, "app"):
         # Kernel is already running a Qt event loop, so there's no need to
         # create another app for it.
         return
@@ -535,13 +562,10 @@ def make_qt_app_for_kernel(gui, kernel):
     set_qt_api_env_from_gui(gui)
 
     # This import is guaranteed to work now:
-    from IPython.external.qt_for_kernel import QtCore, QtGui
+    from IPython.external.qt_for_kernel import QtCore
     from IPython.lib.guisupport import get_app_qt4
 
     kernel.app = get_app_qt4([" "])
-    if isinstance(kernel.app, QtGui.QApplication):
-        kernel.app.setQuitOnLastWindowClosed(False)
-
     kernel.app.qt_event_loop = QtCore.QEventLoop(kernel.app)
 
 
@@ -561,16 +585,22 @@ def enable_gui(gui, kernel=None):
             raise RuntimeError(msg)
     if gui is None:
         # User wants to turn off integration; clear any evidence if Qt was the last one.
-        if hasattr(kernel, 'app'):
-            delattr(kernel, 'app')
+        if hasattr(kernel, "app"):
+            delattr(kernel, "app")
+        if hasattr(kernel, "_qt_notifier"):
+            delattr(kernel, "_qt_notifier")
+        if hasattr(kernel, "_qt_timer"):
+            delattr(kernel, "_qt_timer")
     else:
-        if gui.startswith('qt'):
+        if gui.startswith("qt"):
             # Prepare the kernel here so any exceptions are displayed in the client.
             make_qt_app_for_kernel(gui, kernel)
 
     loop = loop_map[gui]
-    if loop and kernel.eventloop is not None and kernel.eventloop is not loop:
-        msg = "Cannot activate multiple GUI eventloops"
+    if (
+        loop and kernel.eventloop is not None and kernel.eventloop is not loop  # type:ignore[unreachable]
+    ):
+        msg = "Cannot activate multiple GUI eventloops"  # type:ignore[unreachable]
         raise RuntimeError(msg)
     kernel.eventloop = loop
     # We set `eventloop`; the function the user chose is executed in `Kernel.enter_eventloop`, thus
